ENV_DASK_CLUSTER = "DASK_CLUSTER"
ENV_DASK_CACHE_SIZE = "DASK_CACHE_SIZE"
DEFAULT_DASK_CACHE_SIZE = 2e9

ENV_ES_HOSTS = "ES_HOSTS"

ES_HOST = "ES_HOST"


BIOME_EXPLORE_ENDPOINT = "BIOME_EXPLORE_ENDPOINT"

CUDA_DEVICE = "CUDA_DEVICE"
